"""Test that the computations on this branch agree with those on master."""

import pickle
import warnings

import numpy as np
import pytest

from desc.coils import FourierPlanarCoil, FourierRZCoil, FourierXYZCoil, SplineXYZCoil
from desc.compute import data_index, xyz2rpz, xyz2rpz_vec
from desc.examples import get
from desc.geometry import (
    FourierPlanarCurve,
    FourierRZCurve,
    FourierRZToroidalSurface,
    FourierXYZCurve,
    ZernikeRZToroidalSection,
)
from desc.grid import LinearGrid
from desc.magnetic_fields import (
    CurrentPotentialField,
    FourierCurrentPotentialField,
    OmnigenousField,
)
from desc.utils import ResolutionWarning, errorif


def _compare_against_master(
    p, data, master_data, error=False, update_master_data=False
):

    for name in data[p]:
        if p in master_data and name in master_data[p]:
            if np.isnan(master_data[p][name]).all():
                mean = 1.0
            else:
                mean = np.nanmean(np.atleast_1d(np.abs(master_data[p][name])))
            try:
                np.testing.assert_allclose(
                    actual=data[p][name],
                    desired=master_data[p][name],
                    atol=1e-10 * mean + 1e-10,  # add 1e-10 for basically-zero things
                    rtol=1e-10,
                    err_msg=f"Parameterization: {p}. Name: {name}.",
                )
            except AssertionError as e:
                error = True
                print(e)
        else:  # update master data with new compute quantity
            update_master_data = True

    return error, update_master_data


def _xyz_to_rpz(data, name):
    if name in ["x", "center"]:
        res = xyz2rpz(data[name])
    else:
        res = xyz2rpz_vec(data[name], phi=data["phi"])
    return res


def _compare_against_rpz(p, data, data_rpz, coordinate_conversion_func):
    for name in data:
        if data_index[p][name]["dim"] != 3:
            continue
        res = coordinate_conversion_func(data, name) - data_rpz[name]
        errorif(
            not np.all(
                (
                    np.isclose(res, 0, rtol=1e-8, atol=1e-8)
                    | np.isclose(np.abs(res[:, 1]), 2 * np.pi, rtol=1e-8, atol=1e-8)[
                        :, np.newaxis
                    ]
                )
                | (~np.isfinite(data_rpz[name]))
            ),
            AssertionError,
            msg=f"Parameterization: {p}. Name: {name}. Residual {res}",
        )


@pytest.mark.unit
def test_compute_everything():
    """Test that the computations on this branch agree with those on master.

    Also make sure we can compute everything without errors.

    Notes
    -----
    This test will fail if the benchmark file has been updated on both
    the local and upstream branches and git cannot resolve the merge
    conflict. In that case, please regenerate the benchmark file.
    Here are instructions for convenience.

    1. Prepend true to the line near the end of this test.
        ``if True or (not error_rpz and update_master_data_rpz):``
    2. Run pytest -k test_compute_everything
    3. Revert 1.
    4. git add tests/inputs/master_compute_data_rpz.pkl

    """
    elliptic_cross_section_with_torsion = {
        "R_lmn": [10, 1, 0.2],
        "Z_lmn": [-2, -0.2],
        "modes_R": [[0, 0], [1, 0], [0, 1]],
        "modes_Z": [[-1, 0], [0, -1]],
    }
    things = {
        # equilibria
        "desc.equilibrium.equilibrium.Equilibrium": get("W7-X"),
        # curves
        "desc.geometry.curve.FourierXYZCurve": FourierXYZCurve(
            X_n=[5, 10, 2], Y_n=[1, 2, 3], Z_n=[-4, -5, -6]
        ),
        "desc.geometry.curve.FourierRZCurve": FourierRZCurve(
            R_n=[10, 1, 0.2], Z_n=[-2, -0.2], modes_R=[0, 1, 2], modes_Z=[-1, -2], NFP=2
        ),
        "desc.geometry.curve.FourierPlanarCurve": FourierPlanarCurve(
            center=[10, 1, 3], normal=[1, 2, 3], r_n=[1, 2, 3], modes=[0, 1, 2]
        ),
        "desc.geometry.curve.SplineXYZCurve": FourierXYZCurve(
            X_n=[5, 10, 2], Y_n=[1, 2, 3], Z_n=[-4, -5, -6]
        ).to_SplineXYZ(grid=LinearGrid(N=50)),
        # surfaces
        "desc.geometry.surface.FourierRZToroidalSurface": FourierRZToroidalSurface(
            **elliptic_cross_section_with_torsion
        ),
        "desc.geometry.surface.ZernikeRZToroidalSection": ZernikeRZToroidalSection(
            **elliptic_cross_section_with_torsion
        ),
        # magnetic fields
        "desc.magnetic_fields._current_potential.CurrentPotentialField": CurrentPotentialField(  # noqa:E501
            **elliptic_cross_section_with_torsion,
            potential=lambda theta, zeta, G: G * zeta / 2 / np.pi,
            potential_dtheta=lambda theta, zeta, G: np.zeros_like(theta),
            potential_dzeta=lambda theta, zeta, G: G * np.ones_like(theta) / 2 / np.pi,
            params={"G": 1e7},
        ),
        "desc.magnetic_fields._current_potential.FourierCurrentPotentialField": (
            FourierCurrentPotentialField(
                **elliptic_cross_section_with_torsion, I=0, G=1e7
            )
        ),
        "desc.magnetic_fields._core.OmnigenousField": OmnigenousField(
            L_B=0,
            M_B=4,
            L_x=0,
            M_x=1,
            N_x=1,
            NFP=2,
            helicity=(0, 2),
            B_lm=np.array([0.8, 0.9, 1.1, 1.2]),
            x_lmn=np.array([0, -np.pi / 8, 0, np.pi / 8, 0, np.pi / 4]),
        ),
        # coils
        "desc.coils.FourierRZCoil": FourierRZCoil(
            R_n=[10, 1, 0.2], Z_n=[-2, -0.2], modes_R=[0, 1, 2], modes_Z=[-1, -2], NFP=2
        ),
        "desc.coils.FourierXYZCoil": FourierXYZCoil(
            X_n=[5, 10, 2], Y_n=[1, 2, 3], Z_n=[-4, -5, -6]
        ),
        "desc.coils.FourierPlanarCoil": FourierPlanarCoil(
            current=5,
            center=[10, 1, 3],
            normal=[1, 2, 3],
            r_n=[1, 2, 3],
            modes=[0, 1, 2],
        ),
        "desc.coils.SplineXYZCoil": SplineXYZCoil(
            current=5, X=[5, 10, 2, 5], Y=[1, 2, 3, 1], Z=[-4, -5, -6, -4]
        ),
    }
    assert things.keys() == data_index.keys(), (
        f"Missing the parameterization {data_index.keys() - things.keys()}"
        f" to test against master."
    )
    # use this low resolution grid for equilibria to reduce file size
    eqgrid = LinearGrid(
        L=9,
        M=5,
        N=5,
        NFP=things["desc.equilibrium.equilibrium.Equilibrium"].NFP,
        sym=things["desc.equilibrium.equilibrium.Equilibrium"].sym,
        axis=True,
    )
    curvegrid1 = LinearGrid(N=10)
    curvegrid2 = LinearGrid(N=10, NFP=2)
    fieldgrid = LinearGrid(
        L=2,
        M=4,
        N=5,
        NFP=things["desc.magnetic_fields._core.OmnigenousField"].NFP,
        sym=False,
        axis=True,
    )
    grid = {
        "desc.equilibrium.equilibrium.Equilibrium": {"grid": eqgrid},
        "desc.geometry.curve.FourierXYZCurve": {"grid": curvegrid1},
        "desc.geometry.curve.FourierRZCurve": {"grid": curvegrid2},
        "desc.geometry.curve.FourierPlanarCurve": {"grid": curvegrid1},
        "desc.geometry.curve.SplineXYZCurve": {"grid": curvegrid1},
        "desc.magnetic_fields._core.OmnigenousField": {"grid": fieldgrid},
    }

    with open("tests/inputs/master_compute_data_rpz.pkl", "rb") as file:
        master_data_rpz = pickle.load(file)
    this_branch_data_rpz = {}
    update_master_data_rpz = False
    error_rpz = False

    # some things can't compute "phi" and therefore can't convert to XYZ basis
    no_xyz_things = ["desc.magnetic_fields._core.OmnigenousField"]

    with warnings.catch_warnings():
        # Max resolution of master_compute_data.pkl limited by GitHub file
        # size cap at 100 mb, so can't hit suggested resolution for some things.
        warnings.filterwarnings("ignore", category=ResolutionWarning)
        for p in things:
            names = {
                name
                for name in data_index[p]
                # Skip these quantities as they should be covered in other tests.
                if not data_index[p][name]["source_grid_requirement"]
            }
            this_branch_data_rpz[p] = things[p].compute(
                list(names), **grid.get(p, {}), basis="rpz"
            )
            # make sure we can compute everything
            assert this_branch_data_rpz[p].keys() == names, (
                f"Parameterization: {p}. Can't compute "
                + f"{names - this_branch_data_rpz[p].keys()}."
            )
            # compare data against master branch
            error_rpz, update_master_data_rpz = _compare_against_master(
                p,
                this_branch_data_rpz,
                master_data_rpz,
                error_rpz,
                update_master_data_rpz,
            )

            # test compute in XYZ basis
            if p in no_xyz_things:
                continue
            # remove quantities that are not implemented in the XYZ basis
            # TODO (#1110): generalize this instead of hard-coding for
            #  the quantities "grad(B)" & dependencies
            names_xyz = (
                names - {"grad(B)", "|grad(B)|", "L_grad(B)"}
                if "grad(B)" in names
                else names
            )
            this_branch_data_xyz = things[p].compute(
                list(names_xyz), **grid.get(p, {}), basis="xyz"
            )
            assert this_branch_data_xyz.keys() == names_xyz, (
                f"Parameterization: {p}. Can't compute "
                + f"{names_xyz - this_branch_data_xyz.keys()}."
            )
            _compare_against_rpz(
                p, this_branch_data_xyz, this_branch_data_rpz[p], _xyz_to_rpz
            )

    if not error_rpz and update_master_data_rpz:
        # then update the master compute data
        with open("tests/inputs/master_compute_data_rpz.pkl", "wb") as file:
            # remember to git commit this file
            pickle.dump(this_branch_data_rpz, file)

    assert not error_rpz



"""Functions for mapping between flux, sfl, and real space coordinates."""

import functools

import numpy as np

from desc.backend import fori_loop, jit, jnp, put, root, root_scalar, vmap
from desc.compute import compute as compute_fun
from desc.compute import data_index, get_data_deps, get_profiles, get_transforms
from desc.grid import ConcentricGrid, Grid, LinearGrid, QuadratureGrid
from desc.transform import Transform
from desc.utils import check_posint, errorif, safenorm, setdefault, warnif


def _periodic(x, period):
    return jnp.where(jnp.isfinite(period), x % period, x)


def _fixup_residual(r, period):
    r = _periodic(r, period)
    # r should be between -period and period
    return jnp.where((r > period / 2) & jnp.isfinite(period), -period + r, r)


def map_coordinates(  # noqa: C901
    eq,
    coords,
    inbasis,
    outbasis=("rho", "theta", "zeta"),
    guess=None,
    params=None,
    period=None,
    tol=1e-6,
    maxiter=30,
    full_output=False,
    **kwargs,
):
    """Transform coordinates given in ``inbasis`` to ``outbasis``.

    Solves for the computational coordinates that correspond to ``inbasis``,
    then evaluates ``outbasis`` at those locations.

    Performance can often improve significantly given a reasonable initial guess.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium to use.
    coords : ndarray
        Shape (k, 3).
        2D array of input coordinates. Each row is a different point in space.
    inbasis, outbasis : tuple of str
        Labels for input and output coordinates, e.g. ("R", "phi", "Z") or
        ("rho", "alpha", "zeta") or any combination thereof. Labels should be the
        same as the compute function data key.
    guess : jnp.ndarray
        Shape (k, 3).
        Initial guess for the computational coordinates ['rho', 'theta', 'zeta']
        corresponding to ``coords`` in ``inbasis``. If not given, then heuristics
        based on ``inbasis`` or a nearest neighbor search on a grid may be used.
        In general, this must be given to be compatible with JIT.
    params : dict
        Values of equilibrium parameters to use, e.g. ``eq.params_dict``.
    period : tuple of float
        Assumed periodicity for each quantity in ``inbasis``.
        Use ``np.inf`` to denote no periodicity.
    tol : float
        Stopping tolerance.
    maxiter : int
        Maximum number of Newton iterations.
    full_output : bool, optional
        If True, also return a tuple where the first element is the residual from
        the root finding and the second is the number of iterations.
    kwargs : dict, optional
        Additional keyword arguments to pass to ``root`` such as ``maxiter_ls``,
        ``alpha``.

    Returns
    -------
    out : jnp.ndarray
        Shape (k, 3).
        Coordinates mapped from ``inbasis`` to ``outbasis``. Values of NaN will be
        returned for coordinates where root finding did not succeed, possibly
        because the coordinate is not in the plasma volume.
    info : tuple
        2 element tuple containing residuals and number of iterations
        for each point. Only returned if ``full_output`` is True.

    """
    check_posint(maxiter, allow_none=False)
    errorif(
        not np.isfinite(tol) or tol <= 0,
        ValueError,
        f"tol must be a positive float, got {tol}",
    )
    inbasis = tuple(inbasis)
    outbasis = tuple(outbasis)
    params = setdefault(params, eq.params_dict)

    basis_derivs = tuple(f"{X}_{d}" for X in inbasis for d in ("r", "t", "z"))
    for key in basis_derivs:
        errorif(
            key not in data_index["desc.equilibrium.equilibrium.Equilibrium"],
            NotImplementedError,
            f"don't have recipe to compute partial derivative {key}",
        )

    profiles = get_profiles(inbasis + basis_derivs, eq)

    # TODO (#1382): make this work for permutations of in/out basis
    if outbasis == ("rho", "theta", "zeta"):
        if inbasis == ("rho", "alpha", "zeta"):
            if "iota" in kwargs:
                iota = kwargs.pop("iota")
            else:
                if profiles["iota"] is None:
                    profiles["iota"] = eq.get_profile(["iota", "iota_r"], params=params)
                iota = profiles["iota"].compute(Grid(coords, sort=False, jitable=True))
            return _map_clebsch_coordinates(
                coords=coords,
                iota=iota,
                L_lmn=params["L_lmn"],
                L_basis=eq.L_basis,
                guess=guess[:, 1] if guess is not None else None,
                period=period[1] if period is not None else np.inf,
                tol=tol,
                maxiter=maxiter,
                full_output=full_output,
                **kwargs,
            )
        if inbasis == ("rho", "theta_PEST", "zeta"):
            return _map_PEST_coordinates(
                coords=coords,
                L_lmn=params["L_lmn"],
                L_basis=eq.L_basis,
                guess=guess[:, 1] if guess is not None else None,
                period=period[1] if period is not None else np.inf,
                tol=tol,
                maxiter=maxiter,
                full_output=full_output,
                **kwargs,
            )

    # do surface average to get iota once
    if "iota" in profiles and profiles["iota"] is None:
        profiles["iota"] = eq.get_profile(["iota", "iota_r"], params=params)
        params["i_l"] = profiles["iota"].params

    rhomin = kwargs.pop("rhomin", tol / 10)
    period = np.asarray(setdefault(period, (np.inf, np.inf, np.inf)))
    coords = _periodic(coords, period)

    p = "desc.equilibrium.equilibrium.Equilibrium"
    names = inbasis + basis_derivs + outbasis
    deps = list(set(get_data_deps(names, obj=p) + list(names)))

    @functools.partial(jit, static_argnums=1)
    def compute(y, basis):
        grid = Grid(y, sort=False, jitable=True)
        data = {}
        if "iota" in deps:
            data["iota"] = profiles["iota"].compute(grid, params=params["i_l"])
        if "iota_r" in deps:
            data["iota_r"] = profiles["iota"].compute(grid, dr=1, params=params["i_l"])
        if "iota_rr" in deps:
            data["iota_rr"] = profiles["iota"].compute(grid, dr=2, params=params["i_l"])
        transforms = get_transforms(basis, eq, grid, jitable=True)
        data = compute_fun(eq, basis, params, transforms, profiles, data)
        x = jnp.array([data[k] for k in basis]).T
        return x

    @jit
    def residual(y, coords):
        xk = compute(y, inbasis)
        return _fixup_residual(xk - coords, period)

    @jit
    def jac(y, coords):
        J = compute(y, basis_derivs)
        J = J.reshape((3, 3))
        return J

    @jit
    def fixup(y, *args):
        r, t, z = y.T
        # negative rho -> flip theta
        t = jnp.where(r < 0, (t + np.pi), t)
        r = jnp.abs(r)
        r = jnp.clip(r, rhomin, 1)
        y = jnp.array([r, t, z]).T
        return y

    yk = guess
    if yk is None:
        yk = _initial_guess_heuristic(yk, coords, inbasis, eq, profiles)
    if yk is None:
        yk = _initial_guess_nn_search(coords, inbasis, eq, period, compute)

    yk = fixup(yk)

    vecroot = jit(
        vmap(
            lambda x0, *p: root(
                residual,
                x0,
                jac=jac,
                args=p,
                fixup=fixup,
                tol=tol,
                maxiter=maxiter,
                full_output=full_output,
                **kwargs,
            )
        )
    )
    # See description here
    # https://github.com/PlasmaControl/DESC/pull/504#discussion_r1194172532
    # except we make sure properly handle periodic coordinates.
    if full_output:
        yk, (res, niter) = vecroot(yk, coords)
    else:
        yk = vecroot(yk, coords)

    out = compute(yk, outbasis)
    if full_output:
        return out, (res, niter)
    return out


def _initial_guess_heuristic(yk, coords, inbasis, eq, profiles):
    # some qtys have obvious initial guess based on coords
    # commonly, the desired coordinates are something like (radial, poloidal, toroidal)
    radialish = {"rho", "psi"}
    poloidalish = {"theta", "theta_PEST", "alpha"}
    toroidalish = {"zeta", "phi"}

    radial = list(set(inbasis).intersection(radialish))
    poloidal = list(set(inbasis).intersection(poloidalish))
    toroidal = list(set(inbasis).intersection(toroidalish))
    if len(radial) != 1 or len(poloidal) != 1 or len(toroidal) != 1:
        # no heuristics for other cases
        return yk

    rho = theta = zeta = None

    radial = radial[0]
    poloidal = poloidal[0]
    toroidal = toroidal[0]
    if radial == "rho":
        rho = coords[:, inbasis.index("rho")]
    elif radial == "psi":
        rho = jnp.sqrt(coords[:, inbasis.index("psi")] / eq.Psi)

    # omega usually small (zero for now)
    zeta = coords[:, inbasis.index(toroidal)]

    if poloidal == "theta" or poloidal == "theta_PEST":  # lambda usually small
        theta = coords[:, inbasis.index(poloidal)]
    elif poloidal == "alpha":
        alpha = coords[:, inbasis.index("alpha")]
        rho = jnp.atleast_1d(rho)
        zero = jnp.zeros_like(rho)
        grid = Grid(nodes=jnp.column_stack([rho, zero, zero]), sort=False, jitable=True)
        iota = profiles["iota"].compute(grid)
        theta = alpha + iota * zeta

    yk = jnp.column_stack([rho, theta, zeta])
    return yk


def _initial_guess_nn_search(coords, inbasis, eq, period, compute):
    # nearest neighbor search on dense grid
    yg = ConcentricGrid(eq.L_grid, eq.M_grid, max(eq.N_grid, eq.M_grid)).nodes
    xg = compute(yg, inbasis)
    idx = jnp.zeros(len(coords)).astype(int)
    coords = jnp.asarray(coords)

    def _distance_body(i, idx):
        d = _fixup_residual(coords[i] - xg, period)
        distance = safenorm(d, axis=-1)
        k = jnp.argmin(distance)
        idx = put(idx, i, k)
        return idx

    idx = fori_loop(0, len(coords), _distance_body, idx)
    return yg[idx]


# TODO(#568): decide later whether to assume given phi instead of zeta.
def _map_PEST_coordinates(
    coords,
    L_lmn,
    L_basis,
    guess,
    period=np.inf,
    tol=1e-6,
    maxiter=30,
    full_output=False,
    **kwargs,
):
    """Find θ (theta_DESC) for given straight field line ϑ (theta_PEST).

    Parameters
    ----------
    coords : ndarray
        Shape (k, 3).
        Straight field line PEST coordinates [ρ, ϑ, ϕ]. Assumes ζ = ϕ.
        Each row is a different point in space.
    L_lmn : jnp.ndarray
        Spectral coefficients for lambda.
    L_basis : Basis
        Spectral basis for lambda.
    guess : jnp.ndarray
        Shape (k, ).
        Optional initial guess for the computational coordinates.
    period : float
        Assumed periodicity for ϑ.
        Use ``np.inf`` to denote no periodicity.
    tol : float
        Stopping tolerance.
    maxiter : int
        Maximum number of Newton iterations.
    full_output : bool, optional
        If True, also return a tuple where the first element is the residual from
        the root finding and the second is the number of iterations.
    kwargs : dict, optional
        Additional keyword arguments to pass to ``root_scalar`` such as ``maxiter_ls``,
        ``alpha``.

    Returns
    -------
    out : ndarray
        Shape (k, 3).
        DESC computational coordinates [ρ, θ, ζ].
    info : tuple
        2 element tuple containing residuals and number of iterations for each point.
        Only returned if ``full_output`` is True.

    """
    # noqa: D202

    # Root finding for θₖ such that r(θₖ) = ϑₖ(ρ, θₖ, ζ) − ϑ = 0.
    def rootfun(theta, theta_PEST, rho, zeta):
        nodes = jnp.array([rho.squeeze(), theta.squeeze(), zeta.squeeze()], ndmin=2)
        A = L_basis.evaluate(nodes)
        lmbda = A @ L_lmn
        theta_PEST_k = theta + lmbda
        return _fixup_residual(theta_PEST_k - theta_PEST, period).squeeze()

    def jacfun(theta, theta_PEST, rho, zeta):
        # Valid everywhere except θ such that θ+λ = k period where k ∈ ℤ.
        nodes = jnp.array([rho.squeeze(), theta.squeeze(), zeta.squeeze()], ndmin=2)
        A1 = L_basis.evaluate(nodes, (0, 1, 0))
        lmbda_t = jnp.dot(A1, L_lmn)
        return 1 + lmbda_t.squeeze()

    def fixup(x, *args):
        return _periodic(x, period)

    vecroot = jit(
        vmap(
            lambda x0, *p: root_scalar(
                rootfun,
                x0,
                jac=jacfun,
                args=p,
                fixup=fixup,
                tol=tol,
                maxiter=maxiter,
                full_output=full_output,
                **kwargs,
            )
        )
    )
    rho, theta_PEST, zeta = coords.T
    if full_output:
        theta, (res, niter) = vecroot(
            # Assume λ=0 for default initial guess.
            setdefault(guess, theta_PEST),
            theta_PEST,
            rho,
            zeta,
        )
    else:
        theta = vecroot(
            # Assume λ=0 for default initial guess.
            setdefault(guess, theta_PEST),
            theta_PEST,
            rho,
            zeta,
        )
    out = jnp.column_stack([rho, jnp.atleast_1d(theta.squeeze()), zeta])
    if full_output:
        return out, (res, niter)
    return out


# TODO(#568): decide later whether to assume given phi instead of zeta.
def _map_clebsch_coordinates(
    coords,
    iota,
    L_lmn,
    L_basis,
    guess=None,
    period=np.inf,
    tol=1e-6,
    maxiter=30,
    full_output=False,
    **kwargs,
):
    """Find θ for given Clebsch field line poloidal label α.

    Parameters
    ----------
    coords : ndarray
        Shape (k, 3).
        Clebsch field line coordinates [ρ, α, ζ]. Assumes ζ = ϕ.
        Each row is a different point in space.
    iota : ndarray
        Shape (k, ).
        Rotational transform on each node.
    L_lmn : jnp.ndarray
        Spectral coefficients for lambda.
    L_basis : Basis
        Spectral basis for lambda.
    guess : jnp.ndarray
        Shape (k, ).
        Optional initial guess for the computational coordinates.
    period : float
        Assumed periodicity for α.
        Use ``np.inf`` to denote no periodicity.
    tol : float
        Stopping tolerance.
    maxiter : int
        Maximum number of Newton iterations.
    full_output : bool, optional
        If True, also return a tuple where the first element is the residual from
        the root finding and the second is the number of iterations.
    kwargs : dict, optional
        Additional keyword arguments to pass to ``root_scalar`` such as ``maxiter_ls``,
        ``alpha``.

    Returns
    -------
    out : ndarray
        Shape (k, 3).
        DESC computational coordinates [ρ, θ, ζ].
    info : tuple
        2 element tuple containing residuals and number of iterations for each point.
        Only returned if ``full_output`` is True.

    """
    # noqa: D202

    # Root finding for θₖ such that r(θₖ) = αₖ(ρ, θₖ, ζ) − α = 0.
    def rootfun(theta, alpha, rho, zeta, iota):
        nodes = jnp.array([rho.squeeze(), theta.squeeze(), zeta.squeeze()], ndmin=2)
        A = L_basis.evaluate(nodes)
        lmbda = A @ L_lmn
        alpha_k = theta + lmbda - iota * zeta
        return _fixup_residual(alpha_k - alpha, period).squeeze()

    def jacfun(theta, alpha, rho, zeta, iota):
        # Valid everywhere except θ such that θ+λ = k period where k ∈ ℤ.
        nodes = jnp.array([rho.squeeze(), theta.squeeze(), zeta.squeeze()], ndmin=2)
        A1 = L_basis.evaluate(nodes, (0, 1, 0))
        lmbda_t = jnp.dot(A1, L_lmn)
        return 1 + lmbda_t.squeeze()

    def fixup(x, *args):
        return _periodic(x, period)

    vecroot = jit(
        vmap(
            lambda x0, *p: root_scalar(
                rootfun,
                x0,
                jac=jacfun,
                args=p,
                fixup=fixup,
                tol=tol,
                maxiter=maxiter,
                full_output=full_output,
                **kwargs,
            )
        )
    )
    rho, alpha, zeta = coords.T
    if guess is None:
        # Assume λ=0 for default initial guess.
        guess = alpha + iota * zeta
    if full_output:
        theta, (res, niter) = vecroot(guess, alpha, rho, zeta, iota)
    else:
        theta = vecroot(guess, alpha, rho, zeta, iota)

    out = jnp.column_stack([rho, jnp.atleast_1d(theta.squeeze()), zeta])
    if full_output:
        return out, (res, niter)
    return out


def is_nested(eq, grid=None, R_lmn=None, Z_lmn=None, L_lmn=None, msg=None):
    """Check that an equilibrium has properly nested flux surfaces in a plane.

    Does so by checking coordinate Jacobian (sqrt(g)) sign.
    If coordinate Jacobian switches sign somewhere in the volume, this
    indicates that it is zero at some point, meaning surfaces are touching and
    the equilibrium is not nested.

    NOTE: If grid resolution used is too low, or the solution is just barely
    unnested, this function may fail to return the correct answer.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium to use
    grid  :  Grid, optional
        Grid on which to evaluate the coordinate Jacobian and check for the sign.
        (Default to QuadratureGrid with eq's current grid resolutions)
    R_lmn, Z_lmn, L_lmn : ndarray, optional
        spectral coefficients for R, Z, lambda. Defaults to eq.R_lmn, eq.Z_lmn
    msg : {None, "auto", "manual"}
        Warning to throw if unnested.

    Returns
    -------
    is_nested : bool
        whether the surfaces are nested

    """
    if R_lmn is None:
        R_lmn = eq.R_lmn
    if Z_lmn is None:
        Z_lmn = eq.Z_lmn
    if L_lmn is None:
        L_lmn = eq.L_lmn
    if grid is None:
        grid = QuadratureGrid(eq.L_grid, eq.M_grid, eq.N_grid, eq.NFP)

    transforms = get_transforms("sqrt(g)_PEST", obj=eq, grid=grid)
    data = compute_fun(
        "desc.equilibrium.equilibrium.Equilibrium",
        "sqrt(g)_PEST",
        params={"R_lmn": R_lmn, "Z_lmn": Z_lmn, "L_lmn": L_lmn},
        transforms=transforms,
        profiles={},  # no profiles needed
    )

    nested = jnp.all(
        jnp.sign(data["sqrt(g)_PEST"][0]) == jnp.sign(data["sqrt(g)_PEST"])
    )
    warnif(
        not nested and msg is not None,
        RuntimeWarning,
        "Flux surfaces are no longer nested, exiting early. "
        + {
            "auto": "Automatic continuation method failed, consider specifying "
            "continuation steps manually.",
            "manual": "Consider taking smaller perturbation/resolution steps "
            "or reducing trust radius.",
            None: "",
        }[msg],
    )
    return nested


def to_sfl(
    eq,
    L=None,
    M=None,
    N=None,
    L_grid=None,
    M_grid=None,
    N_grid=None,
    rcond=None,
    copy=False,
):
    """Transform this equilibrium to use straight field line coordinates.

    Uses a least squares fit to find FourierZernike coefficients of R, Z, Rb, Zb
    with respect to the straight field line coordinates, rather than the boundary
    coordinates. The new lambda value will be zero.

    NOTE: Though the converted equilibrium will have the same flux surfaces,
    the force balance error will likely be higher than the original equilibrium.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium to use
    L : int, optional
        radial resolution to use for SFL equilibrium. Default = 1.5*eq.L
    M : int, optional
        poloidal resolution to use for SFL equilibrium. Default = 1.5*eq.M
    N : int, optional
        toroidal resolution to use for SFL equilibrium. Default = 1.5*eq.N
    L_grid : int, optional
        radial spatial resolution to use for fit to new basis. Default = 2*L
    M_grid : int, optional
        poloidal spatial resolution to use for fit to new basis. Default = 2*M
    N_grid : int, optional
        toroidal spatial resolution to use for fit to new basis. Default = 2*N
    rcond : float, optional
        cutoff for small singular values in the least squares fit.
    copy : bool, optional
        Whether to update the existing equilibrium or make a copy (Default).

    Returns
    -------
    eq_sfl : Equilibrium
        Equilibrium transformed to a straight field line coordinate representation.

    """
    L = L or int(1.5 * eq.L)
    M = M or int(1.5 * eq.M)
    N = N or int(1.5 * eq.N)
    L_grid = L_grid or int(2 * L)
    M_grid = M_grid or int(2 * M)
    N_grid = N_grid or int(2 * N)

    grid = ConcentricGrid(L_grid, M_grid, N_grid, node_pattern="ocs", NFP=eq.NFP)
    bdry_grid = LinearGrid(M=M, N=N, rho=1.0, NFP=eq.NFP)

    toroidal_coords = eq.compute(["R", "Z", "lambda"], grid=grid)
    theta = grid.nodes[:, 1]
    vartheta = theta + toroidal_coords["lambda"]
    sfl_grid = grid
    sfl_grid.nodes[:, 1] = vartheta

    bdry_coords = eq.compute(["R", "Z", "lambda"], grid=bdry_grid)
    bdry_theta = bdry_grid.nodes[:, 1]
    bdry_vartheta = bdry_theta + bdry_coords["lambda"]
    bdry_sfl_grid = bdry_grid
    bdry_sfl_grid.nodes[:, 1] = bdry_vartheta

    if copy:
        eq_sfl = eq.copy()
    else:
        eq_sfl = eq
    eq_sfl.change_resolution(L, M, N)

    R_sfl_transform = Transform(
        sfl_grid, eq_sfl.R_basis, build=False, build_pinv=True, rcond=rcond
    )
    R_lmn_sfl = R_sfl_transform.fit(toroidal_coords["R"])
    del R_sfl_transform  # these can take up a lot of memory so delete when done.

    Z_sfl_transform = Transform(
        sfl_grid, eq_sfl.Z_basis, build=False, build_pinv=True, rcond=rcond
    )
    Z_lmn_sfl = Z_sfl_transform.fit(toroidal_coords["Z"])
    del Z_sfl_transform
    L_lmn_sfl = np.zeros_like(eq_sfl.L_lmn)

    R_sfl_bdry_transform = Transform(
        bdry_sfl_grid,
        eq_sfl.surface.R_basis,
        build=False,
        build_pinv=True,
        rcond=rcond,
    )
    Rb_lmn_sfl = R_sfl_bdry_transform.fit(bdry_coords["R"])
    del R_sfl_bdry_transform

    Z_sfl_bdry_transform = Transform(
        bdry_sfl_grid,
        eq_sfl.surface.Z_basis,
        build=False,
        build_pinv=True,
        rcond=rcond,
    )
    Zb_lmn_sfl = Z_sfl_bdry_transform.fit(bdry_coords["Z"])
    del Z_sfl_bdry_transform

    eq_sfl.Rb_lmn = Rb_lmn_sfl
    eq_sfl.Zb_lmn = Zb_lmn_sfl
    eq_sfl.R_lmn = R_lmn_sfl
    eq_sfl.Z_lmn = Z_lmn_sfl
    eq_sfl.L_lmn = L_lmn_sfl

    return eq_sfl


def get_rtz_grid(
    eq,
    radial,
    poloidal,
    toroidal,
    coordinates,
    period=(np.inf, np.inf, np.inf),
    jitable=True,
    **kwargs,
):
    """Return DESC grid in (rho, theta, zeta) coordinates from given coordinates.

    Create a tensor-product grid from the given coordinates, and return the same
    grid in DESC coordinates.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium on which to perform coordinate mapping.
    radial : ndarray
        Sorted unique radial coordinates.
    poloidal : ndarray
        Sorted unique poloidal coordinates.
    toroidal : ndarray
        Sorted unique toroidal coordinates.
    coordinates : str
        Input coordinates that are specified by the arguments, respectively.
        raz : rho, alpha, zeta
        rvp : rho, theta_PEST, phi
        rtz : rho, theta, zeta
    period : tuple of float
        Assumed periodicity of the given coordinates.
        Use ``np.inf`` to denote no periodicity.
    jitable : bool, optional
        If false the returned grid has additional attributes.
        Required to be false to retain nodes at magnetic axis.
    kwargs
        Additional parameters to supply to the coordinate mapping function.
        See ``desc.equilibrium.coords.map_coordinates``.

    Returns
    -------
    desc_grid : Grid
        DESC coordinate grid for the given coordinates.

    """
    grid = Grid.create_meshgrid(
        [radial, poloidal, toroidal], coordinates=coordinates, period=period
    )
    if "iota" in kwargs:
        kwargs["iota"] = grid.expand(jnp.atleast_1d(kwargs["iota"]))
    inbasis = {
        "r": "rho",
        "t": "theta",
        "v": "theta_PEST",
        "a": "alpha",
        "z": "zeta",
        "p": "phi",
    }
    rtz_nodes = map_coordinates(
        eq,
        grid.nodes,
        inbasis=[inbasis[char] for char in coordinates],
        outbasis=("rho", "theta", "zeta"),
        period=period,
        **kwargs,
    )
    idx = {}
    if inbasis[coordinates[0]] == "rho":
        # Should work as long as inbasis radial coordinate is
        # single variable, monotonic increasing function of rho.
        idx["_unique_rho_idx"] = grid.unique_rho_idx
        idx["_inverse_rho_idx"] = grid.inverse_rho_idx
    desc_grid = Grid(
        nodes=rtz_nodes,
        coordinates="rtz",
        source_grid=grid,
        sort=False,
        jitable=jitable,
        **idx,
    )
    return desc_grid


# TODO(#1383): deprecated, remove eventually
def compute_theta_coords(
    eq, flux_coords, L_lmn=None, tol=1e-6, maxiter=20, full_output=False, **kwargs
):
    """Find θ (theta_DESC) for given straight field line ϑ (theta_PEST).

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium to use.
    flux_coords : ndarray
        Shape (k, 3).
        Straight field line PEST coordinates [ρ, ϑ, ϕ]. Assumes ζ = ϕ.
        Each row is a different point in space.
    L_lmn : ndarray
        Spectral coefficients for lambda. Defaults to ``eq.L_lmn``.
    tol : float
        Stopping tolerance.
    maxiter : int
        Maximum number of Newton iterations.
    full_output : bool, optional
        If True, also return a tuple where the first element is the residual from
        the root finding and the second is the number of iterations.
    kwargs : dict, optional
        Additional keyword arguments to pass to ``root_scalar`` such as
        ``maxiter_ls``, ``alpha``.

    Returns
    -------
    coords : ndarray
        Shape (k, 3).
        DESC computational coordinates [ρ, θ, ζ].
    info : tuple
        2 element tuple containing residuals and number of iterations for each
        point. Only returned if ``full_output`` is True.

    """
    return eq.compute_theta_coords(
        flux_coords, L_lmn, tol, maxiter, full_output, **kwargs
    )



"""Utilities for getting standard groups of objectives and constraints."""

from desc.utils import flatten_list, is_any_instance, isposint, unique_list

from ._equilibrium import Energy, ForceBalance, HelicalForceBalance, RadialForceBalance
from .linear_objectives import (
    AxisRSelfConsistency,
    AxisZSelfConsistency,
    BoundaryRSelfConsistency,
    BoundaryZSelfConsistency,
    FixAnisotropy,
    FixAtomicNumber,
    FixAxisR,
    FixAxisZ,
    FixBoundaryR,
    FixBoundaryZ,
    FixCurrent,
    FixCurveRotation,
    FixCurveShift,
    FixElectronDensity,
    FixElectronTemperature,
    FixIonTemperature,
    FixIota,
    FixLambdaGauge,
    FixNearAxisLambda,
    FixNearAxisR,
    FixNearAxisZ,
    FixPressure,
    FixPsi,
    FixSheetCurrent,
)
from .nae_utils import calc_zeroth_order_lambda, make_RZ_cons_1st_order
from .objective_funs import ObjectiveFunction

_PROFILE_CONSTRAINTS = {
    "pressure": FixPressure,
    "iota": FixIota,
    "current": FixCurrent,
    "electron_density": FixElectronDensity,
    "electron_temperature": FixElectronTemperature,
    "ion_temperature": FixIonTemperature,
    "atomic_number": FixAtomicNumber,
    "anisotropy": FixAnisotropy,
}


def get_equilibrium_objective(eq, mode="force", normalize=True, jac_chunk_size="auto"):
    """Get the objective function for a typical force balance equilibrium problem.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    mode : one of {"force", "forces", "energy"}
        which objective to return. "force" computes force residuals on unified grid.
        "forces" uses two different grids for radial and helical forces. "energy" is
        for minimizing MHD energy.
    normalize : bool
        Whether to normalize units of objective.
    jac_chunk_size : int or "auto", optional
        If `"batched"` deriv_mode is used, will calculate the Jacobian
        ``jac_chunk_size`` columns at a time, instead of all at once.
        The memory usage of the Jacobian calculation is roughly
        ``memory usage = m0 + m1*jac_chunk_size``: the smaller the chunk size,
        the less memory the Jacobian calculation will require (with some baseline
        memory usage). The time it takes to compute the Jacobian is roughly
        ``t= t0 + t1/jac_chunk_size` so the larger the ``jac_chunk_size``, the faster
        the calculation takes, at the cost of requiring more memory.
        If None, it will use the largest size i.e ``obj.dim_x``.
        Defaults to ``chunk_size="auto"`` which will use a conservative
        chunk size based off of a heuristic estimate of the memory usage.


    Returns
    -------
    objective, ObjectiveFunction
        An objective function with default force balance objectives.

    """
    kwargs = {"eq": eq, "normalize": normalize, "normalize_target": normalize}
    if mode == "energy":
        objectives = Energy(**kwargs)
    elif mode == "force":
        objectives = ForceBalance(**kwargs)
    elif mode == "forces":
        objectives = (RadialForceBalance(**kwargs), HelicalForceBalance(**kwargs))
    else:
        raise ValueError("got an unknown equilibrium objective type '{}'".format(mode))
    deriv_mode = "batched" if isposint(jac_chunk_size) else "auto"
    return ObjectiveFunction(
        objectives, jac_chunk_size=jac_chunk_size, deriv_mode=deriv_mode
    )


def get_fixed_axis_constraints(eq, profiles=True, normalize=True):
    """Get the constraints necessary for a fixed-axis equilibrium problem.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium to constrain.
    profiles : bool
        If True, also include constraints to fix all profiles assigned to equilibrium.
    normalize : bool
        Whether to apply constraints in normalized units.

    Returns
    -------
    constraints, tuple of _Objectives
        A list of the linear constraints used in fixed-axis problems.

    """
    kwargs = {"eq": eq, "normalize": normalize, "normalize_target": normalize}
    constraints = (FixAxisR(**kwargs), FixAxisZ(**kwargs), FixPsi(**kwargs))
    if profiles:
        for name, con in _PROFILE_CONSTRAINTS.items():
            if getattr(eq, name) is not None:
                constraints += (con(**kwargs),)
    constraints += (FixSheetCurrent(**kwargs),)

    return constraints


def get_fixed_boundary_constraints(eq, profiles=True, normalize=True):
    """Get the constraints necessary for a typical fixed-boundary equilibrium problem.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium to constrain.
    profiles : bool
        If True, also include constraints to fix all profiles assigned to equilibrium.
    normalize : bool
        Whether to apply constraints in normalized units.

    Returns
    -------
    constraints, tuple of _Objectives
        A list of the linear constraints used in fixed-boundary problems.

    """
    kwargs = {"eq": eq, "normalize": normalize, "normalize_target": normalize}
    constraints = (FixBoundaryR(**kwargs), FixBoundaryZ(**kwargs), FixPsi(**kwargs))
    if profiles:
        for name, con in _PROFILE_CONSTRAINTS.items():
            if getattr(eq, name) is not None:
                constraints += (con(**kwargs),)
    constraints += (FixSheetCurrent(**kwargs),)

    return constraints


def get_NAE_constraints(
    desc_eq,
    qsc_eq,
    order=1,
    profiles=True,
    normalize=True,
    N=None,
    fix_lambda=False,
):
    """Get the constraints necessary for fixing NAE behavior in an equilibrium problem.

    Parameters
    ----------
    desc_eq : Equilibrium
        Equilibrium to constrain behavior of
        (assumed to be a fit from the NAE equil using `.from_near_axis()`).
    qsc_eq : Qsc, optional
        Qsc object defining the near-axis equilibrium to constrain behavior to.
        if None, will instead fix the current near-axis behavior of the ``desc_eq``
    order : {0,1,2}
        order (in rho) of near-axis behavior to constrain
    profiles : bool
        If True, also include constraints to fix all profiles assigned to equilibrium.
    normalize : bool
        Whether to apply constraints in normalized units.
    N : int
        max toroidal resolution to constrain.
        If `None`, defaults to equilibrium's toroidal resolution
    fix_lambda : bool or int
        Whether to constrain lambda to match that of the NAE near-axis
        if an `int`, fixes lambda up to that order in rho {0,1}
        if `True`, fixes lambda up to the specified order given by `order`

    Returns
    -------
    constraints, tuple of _Objectives
        A list of the linear constraints used in fixed-axis problems.

    """
    kwargs = {"eq": desc_eq, "normalize": normalize, "normalize_target": normalize}
    if not isinstance(fix_lambda, bool):
        fix_lambda = int(fix_lambda)
    constraints = (FixPsi(**kwargs),)

    if profiles:
        for name, con in _PROFILE_CONSTRAINTS.items():
            if getattr(desc_eq, name) is not None:
                constraints += (con(**kwargs),)
    constraints += (FixSheetCurrent(**kwargs),)

    constraints += (
        FixNearAxisR(
            eq=desc_eq,
            target=qsc_eq,
            N=N,
            order=order,
            normalize=normalize,
        ),
        FixNearAxisZ(
            eq=desc_eq,
            target=qsc_eq,
            N=N,
            order=order,
            normalize=normalize,
        ),
    )
    if fix_lambda or (fix_lambda >= 0 and type(fix_lambda) is int):
        constraints += (
            FixNearAxisLambda(
                eq=desc_eq,
                target=qsc_eq,
                N=N,
                order=int(fix_lambda),
                normalize=normalize,
            ),
        )

    return constraints


def _get_NAE_constraints(
    desc_eq,
    qsc_eq,
    order=1,
    N=None,
    fix_lambda=False,
    normalize=True,
):
    """Get the constraints necessary for fixing NAE behavior in an equilibrium problem.

    NOTE: This will return tuples of FixSumModes__, this is not intended to be directly
    used by the user. Instead, call the ``get_NAE_constraints`` function, or use the
    FixNearAxis{R,Z,Lambda} objectives along with the FixAxis{R,Z} objectives. This
    instead is a helper function to get the needed constraints for the
    FixNearAxis{R,Z,Lambda} objectives, which use those constraints to form
    the full constraint matrices to properly constrain the behavior of that
    part of the equilibrium.

    Parameters
    ----------
    desc_eq : Equilibrium
        Equilibrium to constrain behavior of
        (assumed to be a fit from the NAE equil using `.from_near_axis()`).
    qsc_eq : Qsc
        Qsc object defining the near-axis equilibrium to constrain behavior to.
        order : {0,1,2}
        order (in rho) of near-axis behavior to constrain
    normalize : bool
        Whether to apply constraints in normalized units.
    N : int
        max toroidal resolution to constrain.
        If `None`, defaults to equilibrium's toroidal resolution
    fix_lambda : bool or int
        Whether to constrain lambda to match that of the NAE near-axis
        if an `int`, fixes lambda up to that order in rho {0,1}
        if `True`, fixes lambda up to the specified order given by `order`
    normalize : bool
        Whether to apply constraints in normalized units.

    Returns
    -------
    constraints, tuple of _Objectives
        A list of the linear constraints used in fixed-near-axis behavior problems.
    """
    if not isinstance(fix_lambda, bool):
        fix_lambda = int(fix_lambda)
    constraints = ()

    kwargs = {"eq": desc_eq, "normalize": normalize, "normalize_target": normalize}
    if not isinstance(fix_lambda, bool):
        fix_lambda = int(fix_lambda)

    if fix_lambda or (fix_lambda >= 0 and type(fix_lambda) is int):
        L_axis_constraints, _, _ = calc_zeroth_order_lambda(
            qsc=qsc_eq, desc_eq=desc_eq, N=N
        )
        constraints += L_axis_constraints

    # Axis constraints
    constraints += (
        FixAxisR(**kwargs),
        FixAxisZ(**kwargs),
        AxisRSelfConsistency(desc_eq),
        AxisZSelfConsistency(desc_eq),
    )

    if order >= 1:  # first order constraints
        constraints += make_RZ_cons_1st_order(
            qsc=qsc_eq, desc_eq=desc_eq, N=N, fix_lambda=fix_lambda and fix_lambda > 0
        )
    if order >= 2:  # 2nd order constraints
        raise NotImplementedError("NAE constraints only implemented up to O(rho) ")

    return constraints


def maybe_add_self_consistency(thing, constraints):
    """Add self consistency constraints if needed."""
    params = set(unique_list(flatten_list(thing.optimizable_params))[0])

    # Equilibrium
    if {"R_lmn", "Rb_lmn"} <= params and not is_any_instance(
        constraints, BoundaryRSelfConsistency
    ):
        constraints += (BoundaryRSelfConsistency(eq=thing),)
    if {"Z_lmn", "Zb_lmn"} <= params and not is_any_instance(
        constraints, BoundaryZSelfConsistency
    ):
        constraints += (BoundaryZSelfConsistency(eq=thing),)
    if {"L_lmn"} <= params and not is_any_instance(constraints, FixLambdaGauge):
        constraints += (FixLambdaGauge(eq=thing),)
    if {"R_lmn", "Ra_n"} <= params and not is_any_instance(
        constraints, AxisRSelfConsistency
    ):
        constraints += (AxisRSelfConsistency(eq=thing),)
    if {"Z_lmn", "Za_n"} <= params and not is_any_instance(
        constraints, AxisZSelfConsistency
    ):
        constraints += (AxisZSelfConsistency(eq=thing),)

    # Curve
    if {"shift"} <= params and not is_any_instance(constraints, FixCurveShift):
        constraints += (FixCurveShift(curve=thing),)
    if {"rotmat"} <= params and not is_any_instance(constraints, FixCurveRotation):
        constraints += (FixCurveRotation(curve=thing),)

    return constraints



"""Generic objectives that don't belong anywhere else."""

import inspect
import re

import numpy as np

from desc.backend import jnp, tree_flatten, tree_leaves, tree_unflatten
from desc.compute import data_index
from desc.compute.utils import _compute as compute_fun
from desc.compute.utils import _parse_parameterization, get_profiles, get_transforms
from desc.grid import QuadratureGrid
from desc.optimizable import OptimizableCollection
from desc.utils import errorif, parse_argname_change

from .linear_objectives import _FixedObjective
from .objective_funs import _Objective, collect_docs


class GenericObjective(_Objective):
    """A generic objective that can compute any quantity from the `data_index`.

    Parameters
    ----------
    f : str
        Name of the quantity to compute.
    thing : Optimizable
        Object that will be optimized to satisfy the Objective.
    grid : Grid, optional
        Collocation grid containing the nodes to evaluate at. Defaults to
        ``QuadratureGrid(eq.L_grid, eq.M_grid, eq.N_grid)`` if thing is an Equilibrium.

    """

    __doc__ = __doc__.rstrip() + collect_docs(
        target_default="``target=0``.", bounds_default="``target=0``."
    )

    _print_value_fmt = "Generic objective value: "

    def __init__(
        self,
        f,
        thing,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        loss_function=None,
        deriv_mode="auto",
        grid=None,
        name="generic",
        jac_chunk_size=None,
        **kwargs,
    ):
        errorif(
            isinstance(thing, OptimizableCollection),
            NotImplementedError,
            "thing must be of type Optimizable and not OptimizableCollection.",
        )
        thing = parse_argname_change(thing, kwargs, "eq", "thing")
        if target is None and bounds is None:
            target = 0
        self.f = f
        self._grid = grid
        super().__init__(
            things=thing,
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            loss_function=loss_function,
            deriv_mode=deriv_mode,
            name=name,
            jac_chunk_size=jac_chunk_size,
        )
        self._p = _parse_parameterization(thing)
        self._scalar = not bool(data_index[self._p][self.f]["dim"])
        self._coordinates = data_index[self._p][self.f]["coordinates"]
        self._units = "(" + data_index[self._p][self.f]["units"] + ")"

    def build(self, use_jit=True, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        thing = self.things[0]
        if self._grid is None:
            errorif(
                self._p != "desc.equilibrium.equilibrium.Equilibrium",
                ValueError,
                "grid must be supplied for things besides an Equilibrium.",
            )
            grid = QuadratureGrid(thing.L_grid, thing.M_grid, thing.N_grid, thing.NFP)
        else:
            grid = self._grid

        if data_index[self._p][self.f]["dim"] == 0:
            self._dim_f = 1
        elif data_index[self._p][self.f]["coordinates"] == "r":
            self._dim_f = grid.num_rho
        else:
            self._dim_f = grid.num_nodes * np.prod(data_index[self._p][self.f]["dim"])
        profiles = get_profiles(self.f, obj=thing, grid=grid)
        transforms = get_transforms(self.f, obj=thing, grid=grid)
        self._constants = {
            "transforms": transforms,
            "profiles": profiles,
        }
        super().build(use_jit=use_jit, verbose=verbose)

    def compute(self, params, constants=None):
        """Compute the quantity.

        Parameters
        ----------
        params : dict
            Dictionary of equilibrium degrees of freedom, eg Equilibrium.params_dict
        constants : dict
            Dictionary of constant data, eg transforms, profiles etc.
            Defaults to self.constants

        Returns
        -------
        f : ndarray
            Computed quantity.

        """
        if constants is None:
            constants = self.constants
        data = compute_fun(
            self._p,
            self.f,
            params=params,
            transforms=constants["transforms"],
            profiles=constants["profiles"],
        )
        f = data[self.f]
        if self._coordinates == "r":
            f = constants["transforms"]["grid"].compress(f, surface_label="rho")
        return f.flatten(order="F")  # so that default quad weights line up correctly.


class LinearObjectiveFromUser(_FixedObjective):
    """Wrap a user defined linear objective function.

    The user supplied function should take one argument, ``params``, which is a
    dictionary of parameters of an Optimizable "thing".

    The function should be JAX traceable and differentiable, and should return a single
    JAX array. The source code of the function must be visible to the ``inspect`` module
    for parsing.

    Parameters
    ----------
    fun : callable
        Custom objective function.
    thing : Optimizable
        Object whose degrees of freedom are being constrained.

    """

    __doc__ = __doc__.rstrip() + collect_docs(
        target_default="``target=0``.", bounds_default="``target=0``."
    )

    _scalar = False
    _linear = True
    _fixed = True
    _units = "(Unknown)"
    _print_value_fmt = "Custom linear objective value: "

    def __init__(
        self,
        fun,
        thing,
        target=None,
        bounds=None,
        weight=1,
        normalize=False,
        normalize_target=False,
        name="custom linear",
        jac_chunk_size=None,
    ):
        if target is None and bounds is None:
            target = 0
        self._fun = fun
        super().__init__(
            things=thing,
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
            jac_chunk_size=jac_chunk_size,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        thing = self.things[0]

        import jax

        self._dim_f = jax.eval_shape(self._fun, thing.params_dict).size

        # check that fun is linear
        params, params_tree = tree_flatten(thing.params_dict)
        for param in params:
            param += np.random.rand(param.size) * 10
        params = tree_unflatten(params_tree, params)
        J1 = jax.jacrev(self._fun)(thing.params_dict)
        J2 = jax.jacrev(self._fun)(params)
        for j1, j2 in zip(tree_leaves(J1), tree_leaves(J2)):
            assert np.all(j1 == j2), "Function must be linear!"

        super().build(use_jit=use_jit, verbose=verbose)

    def compute(self, params, constants=None):
        """Compute fixed degree of freedom errors.

        Parameters
        ----------
        params : dict
            Dictionary of equilibrium degrees of freedom, eg thing.params_dict
        constants : dict
            Dictionary of constant data, eg transforms, profiles etc. Defaults to
            self.constants

        Returns
        -------
        f : ndarray
            Fixed degree of freedom errors.

        """
        f = self._fun(params)
        return f


class ObjectiveFromUser(_Objective):
    """Wrap a user defined objective function.

    The user supplied function should take two arguments: ``grid`` and ``data``.

    ``grid`` is the Grid object containing the nodes where the data is computed.

    ``data`` is a dictionary of values with keys from the list of `variables`_. Values
    will be the given data evaluated at ``grid``.

    The function should be JAX traceable and differentiable, and should return a single
    JAX array. The source code of the function must be visible to the ``inspect`` module
    for parsing.

    .. _variables: https://desc-docs.readthedocs.io/en/stable/variables.html

    Parameters
    ----------
    fun : callable
        Custom objective function.
    thing : Optimizable
        Object that will be optimized to satisfy the Objective.
    grid : Grid, optional
        Collocation grid containing the nodes to evaluate at. Defaults to
        ``QuadratureGrid(eq.L_grid, eq.M_grid, eq.N_grid)`` if thing is an Equilibrium.

    """

    __doc__ = __doc__.rstrip() + collect_docs(
        target_default="``target=0``.", bounds_default="``target=0``."
    )
    __doc__ += """
    Examples
    --------
    .. code-block:: python

        from desc.compute.utils import surface_averages

        def myfun(grid, data):
            # This will compute the flux surface average of the function
            # R*B_T from the Grad-Shafranov equation
            f = data['R']*data['B_phi']
            f_fsa = surface_averages(grid, f, sqrt_g=data['sqrt_g'])
            # this has the FSA values on the full grid, but we just want
            # the unique values:
            return grid.compress(f_fsa)

        myobj = ObjectiveFromUser(fun=myfun, thing=eq)

    """

    _units = "(Unknown)"
    _print_value_fmt = "Custom objective value: "

    def __init__(
        self,
        fun,
        thing,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        loss_function=None,
        deriv_mode="auto",
        grid=None,
        name="custom",
        jac_chunk_size=None,
        **kwargs,
    ):
        errorif(
            isinstance(thing, OptimizableCollection),
            NotImplementedError,
            "thing must be of type Optimizable and not OptimizableCollection.",
        )
        thing = parse_argname_change(thing, kwargs, "eq", "thing")
        if target is None and bounds is None:
            target = 0
        self._fun = fun
        self._grid = grid
        super().__init__(
            things=thing,
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            loss_function=loss_function,
            deriv_mode=deriv_mode,
            name=name,
            jac_chunk_size=jac_chunk_size,
        )
        self._p = _parse_parameterization(thing)

    def build(self, use_jit=True, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        thing = self.things[0]
        if self._grid is None:
            errorif(
                self._p != "desc.equilibrium.equilibrium.Equilibrium",
                ValueError,
                "grid must be supplied for things besides an Equilibrium.",
            )
            grid = QuadratureGrid(thing.L_grid, thing.M_grid, thing.N_grid, thing.NFP)
        else:
            grid = self._grid

        def get_vars(fun):
            pattern = r"data\[(.*?)\]"
            src = inspect.getsource(fun)
            variables = re.findall(pattern, src)
            variables = list({s.strip().strip("'").strip('"') for s in variables})
            return variables

        self._data_keys = get_vars(self._fun)
        dummy_data = {}
        for key in self._data_keys:
            assert key in data_index[self._p], f"Don't know how to compute {key}."
            if data_index[self._p][key]["dim"] == 0:
                dummy_data[key] = jnp.array(0.0)
            else:
                dummy_data[key] = jnp.empty(
                    (grid.num_nodes, data_index[self._p][key]["dim"])
                ).squeeze()

        self._fun_wrapped = lambda data: self._fun(grid, data)
        import jax

        self._dim_f = jax.eval_shape(self._fun_wrapped, dummy_data).size
        self._scalar = self._dim_f == 1
        profiles = get_profiles(self._data_keys, obj=thing, grid=grid)
        transforms = get_transforms(self._data_keys, obj=thing, grid=grid)
        self._constants = {
            "transforms": transforms,
            "profiles": profiles,
            "quad_weights": 1.0,
        }

        super().build(use_jit=use_jit, verbose=verbose)

    def compute(self, params, constants=None):
        """Compute the quantity.

        Parameters
        ----------
        params : dict
            Dictionary of equilibrium degrees of freedom, eg Equilibrium.params_dict
        constants : dict
            Dictionary of constant data, eg transforms, profiles etc. Defaults to
            self.constants

        Returns
        -------
        f : ndarray
            Computed quantity.

        """
        if constants is None:
            constants = self.constants
        data = compute_fun(
            self._p,
            self._data_keys,
            params=params,
            transforms=constants["transforms"],
            profiles=constants["profiles"],
        )
        f = self._fun_wrapped(data)
        return f
